import os
from flask import request
from flask_restx import Resource, abort
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.util import ns_conf
from mindsdb import __about__

TELEMETRY_FILE = 'telemetry.lock'

@ns_conf.route('/ping')
class Ping(Resource):
    @ns_conf.doc('get_ping')
    def get(self):
        '''Checks server avaliable'''
        return {'status': 'ok'}

@ns_conf.route('/report_uuid')
class ReportUUID(Resource):
    @ns_conf.doc('get_report_uuid')
    def get(self):
        metamodel_name = '___monitroing_metamodel___'
        predictor = ca.mindsdb_native.create(metamodel_name)
        return {
            'report_uuid': predictor.report_uuid
        }

@ns_conf.route('/telemetry')
class Telemetry(Resource):
    @ns_conf.doc('get_telemetry_status')
    def get(self):
        status = "enabled" if is_telemetry_active() else "disabled"
        return {"status": status}

    @ns_conf.doc('set_telemetry')
    def post(self):
        data = request.json
        action = data['action']
        if str(action).lower() in ["true", "enable", "on"]:
            enable_telemetry()
        else:
            disable_telemetry()


def enable_telemetry():
    path = os.path.join(ca.config_obj['storage_dir'], TELEMETRY_FILE)
    if os.path.exists(path):
        os.remove(path)

def disable_telemetry():
    path = os.path.join(ca.config_obj['storage_dir'], TELEMETRY_FILE)
    with open(path, 'w') as _:
        pass

def is_telemetry_active():
    path = os.path.join(ca.config_obj['storage_dir'], TELEMETRY_FILE)
    return not os.path.exists(path)



import os
import json
from datetime import datetime, timedelta


def _in(ask, default, use_default):
    if use_default:
        return default

    user_input = input(f'{ask} (Default: {default})')
    if user_input is None or user_input == '':
        user_input = default

    if type(default) == int:
        user_input = int(user_input)

    if type(default) == bool:
        user_input = int(user_input)

    return user_input


def auto_config(python_path, pip_path, storage_dir):
    config = {
        "debug": False,
        "config_version": "1.4",
        "api": {
        },
        "integrations": {
            "default_clickhouse": {
                "publish": False,
                "type": 'clickhouse'
            },
            "default_mariadb": {
                "publish": False,
                "type": 'mariadb'
            },
            "default_mysql": {
                "publish": False,
                "type": 'mysql'
            },
            "default_postgres": {
                "publish": False,
                "type": 'postgres',
                "database": 'postgres'
            },
            "default_mssql": {
                "publish": False,
                "type": 'mssql'
            },
            "default_mongodb": {
                "publish": False,
                "type": 'mongodb'
            }
        },
        'storage_dir': storage_dir
    }

    if isinstance(python_path, str):
        config['python_interpreter'] = python_path

    if isinstance(pip_path, str):
        config['pip_path'] = python_path

    return config


def make_ssl_cert(file_path):
    from cryptography import x509
    from cryptography.x509.oid import NameOID
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, 'mdb_autogen'),
        x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'California'),
        x509.NameAttribute(NameOID.LOCALITY_NAME, 'Berkeley'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'MindsDB')
    ])

    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(key.public_key())
        .serial_number(1)
        .not_valid_before(now - timedelta(days=10 * 365))
        .not_valid_after(now + timedelta(days=10 * 365))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=0),
            False
        )
        .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    with open(file_path, 'wb') as f:
        f.write(key_pem + cert_pem)


def cli_config(python_path, pip_path, storage_dir, config_dir, use_default=False):
    config = auto_config(python_path, pip_path, storage_dir)

    http = _in('Enable HTTP API ? [Y/N]', 'Y', use_default)
    if http in ['Y', 'y']:
        config['api']['http'] = {}
        config['api']['http']['host'] = _in('HTTP interface host: ', '127.0.0.1', use_default)
        config['api']['http']['port'] = _in('HTTP interface port: ', '47334', use_default)

    mysql = _in('Enable MYSQL API ? [Y/N]', 'Y', use_default)
    if mysql in ['Y', 'y']:
        config['api']['mysql'] = {}
        config['api']['mysql']['host'] = _in('MYSQL interface host', '127.0.0.1', use_default)
        config['api']['mysql']['port'] = _in('MYSQL interface port', '47335', use_default)
        config['api']['mysql']['user'] = _in('MYSQL interface user', 'mindsdb', use_default)
        config['api']['mysql']['password'] = _in('MYSQL interface password', '', use_default)

    mongodb = _in('Enable Mongo API ? [Y/N]', 'Y', use_default)
    if mongodb in ['Y', 'y']:
        config['api']['mongodb'] = {}
        config['api']['mongodb']['host'] = _in('Mongo interface host: ', '127.0.0.1', use_default)
        config['api']['mongodb']['port'] = _in('Mongo interface port: ', '47336', use_default)

    clickhouse = _in('Connect to clickhouse ? [Y/N]', 'Y', use_default)
    if clickhouse in ['Y', 'y']:
        config['integrations']['default_clickhouse']['publish'] = _in('Enable Clickhouse integration?: ', False, use_default)
        config['integrations']['default_clickhouse']['host'] = _in('Clickhouse host: ', '127.0.0.1', use_default)
        config['integrations']['default_clickhouse']['port'] = _in('Clickhouse port: ', 8123, use_default)
        config['integrations']['default_clickhouse']['user'] = _in('Clickhouse user: ', 'default', use_default)
        config['integrations']['default_clickhouse']['password'] = _in('Clickhouse password: ', '', use_default)
        config['integrations']['default_clickhouse']['type'] = 'clickhouse'

    mariadb = _in('Connect to Mariadb ? [Y/N]', 'Y', use_default)
    if mariadb in ['Y', 'y']:
        config['integrations']['default_mariadb']['publish'] = _in('Enable Mariadb integration?: ', False, use_default)
        config['integrations']['default_mariadb']['host'] = _in('Mariadb host: ', '127.0.0.1', use_default)
        config['integrations']['default_mariadb']['port'] = _in('Mariadb port: ', 3306, use_default)
        config['integrations']['default_mariadb']['user'] = _in('Mariadb user: ', 'root', use_default)
        config['integrations']['default_mariadb']['password'] = _in('Mariadb password: ', '', use_default)
        config['integrations']['default_mariadb']['type'] = 'mariadb'

    mysql = _in('Connect to MySQL ? [Y/N]', 'Y', use_default)
    if mysql in ['Y', 'y']:
        config['integrations']['default_mysql']['publish'] = _in('Enable MySQL integration?: ', False, use_default)
        config['integrations']['default_mysql']['host'] = _in('MySQL host: ', '127.0.0.1', use_default)
        config['integrations']['default_mysql']['port'] = _in('MySQL port: ', 3306, use_default)
        config['integrations']['default_mysql']['user'] = _in('MySQL user: ', 'root', use_default)
        config['integrations']['default_mysql']['password'] = _in('MySQL password: ', '', use_default)
        config['integrations']['default_mysql']['type'] = 'mysql'

    mysql = _in('Connect to PostgreSQL ? [Y/N]', 'Y', use_default)
    if mysql in ['Y', 'y']:
        config['integrations']['default_postgres']['publish'] = _in('Enable PostgreSQL integration?: ', False, use_default)
        config['integrations']['default_postgres']['host'] = _in('PostgreSQL host: ', '127.0.0.1', use_default)
        config['integrations']['default_postgres']['port'] = _in('PostgreSQL port: ', 5432, use_default)
        config['integrations']['default_postgres']['user'] = _in('PostgreSQL user: ', 'postgres', use_default)
        config['integrations']['default_postgres']['password'] = _in('PostgreSQL password: ', '', use_default)
        config['integrations']['default_postgres']['database'] = _in('PostgreSQL database: ', 'postgres', use_default)
        config['integrations']['default_postgres']['type'] = 'postgres'

    mssql = _in('Connect to MSSQL ? [Y/N]', 'Y', use_default)
    if mssql in ['Y', 'y']:
        config['integrations']['default_mssql']['publish'] = _in('Enable MSSQL integration?: ', False, use_default)
        config['integrations']['default_mssql']['host'] = _in('MSSQL host: ', '127.0.0.1', use_default)
        config['integrations']['default_mssql']['port'] = _in('MSSQL port: ', 1433, use_default)
        config['integrations']['default_mssql']['user'] = _in('MSSQL user: ', 'sa', use_default)
        config['integrations']['default_mssql']['password'] = _in('MSSQL password: ', '', use_default)
        config['integrations']['default_mssql']['odbc_driver_name'] = _in('MySQL ODBC driver name: ', 'MySQL ODBC 8.0 Unicode Driver', use_default)
        config['integrations']['default_mssql']['type'] = 'mssql'

    mongodb = _in('Connect to MongoDB ? [Y/N]', 'Y', use_default)
    if mongodb in ['Y', 'y']:
        config['integrations']['default_mongodb']['publish'] = _in('Enable MongoDB integration?: ', False, use_default)
        config['integrations']['default_mongodb']['host'] = _in('MongoDB host: ', '127.0.0.1', use_default)
        config['integrations']['default_mongodb']['port'] = _in('MongoDB port: ', 27017, use_default)
        config['integrations']['default_mongodb']['user'] = _in('MongoDB user: ', '', use_default)
        config['integrations']['default_mongodb']['password'] = _in('MongoDB password: ', '', use_default)
        config['integrations']['default_mongodb']['type'] = 'mongodb'

    for db_name in list(config['integrations'].keys()):
        if not config['integrations'][db_name]['publish']:
            del config['integrations'][db_name]

    config_path = os.path.join(config_dir, 'config.json')
    with open(config_path, 'w') as fp:
        json.dump(config, fp, indent=4, sort_keys=True)

    return config_path


def daemon_creator(python_path, config_path=None):
    daemon_path = '/etc/systemd/system/mindsdb.service'
    service_txt = f"""[Unit]
        Description=Mindsdb
        [Service]
        ExecStart={python_path} -m mindsdb { "--config="+config_path  if config_path else ""}
        [Install]
        WantedBy=multi-user.target
    """.strip(' ')

    try:
        with open(daemon_path, 'w') as fp:
            fp.write(service_txt)
    except Exception as e:
        print(f'Failed to create daemon, error: {e}')

    try:
        os.system('systemctl daemon-reload')
    except Exception as e:
        print(f'Failed to load daemon, error: {e}')
    return daemon_path


def make_executable(python_path, exec_path, config_path=None, update=False):
    text = f"""#!/bin/bash
update={str(update).lower()}

if [ "$update" = true ] 
    then
    {python_path} -m pip install mindsdb --upgrade 
fi
{python_path} -m mindsdb { "--config="+config_path  if config_path else ""}
"""

    with open(exec_path, 'w') as fp:
        fp.write(text)

    os.system(f'chmod +x {exec_path}')



import torch.multiprocessing as mp
from mindsdb.interfaces.database.database import DatabaseWrapper


ctx = mp.get_context('spawn')


class PredictorProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(PredictorProcess, self).__init__(args=args)

    def run(self):
        '''
        running at subprocess due to
        ValueError: signal only works in main thread

        this is work for celery worker here?
        '''
        import mindsdb_native

        name, from_data, to_predict, kwargs, config, trx_type = self._args

        mdb = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})

        if trx_type == 'learn':
            to_predict = to_predict if isinstance(to_predict, list) else [to_predict]
            data_source = getattr(mindsdb_native, from_data['class'])(*from_data['args'], **from_data['kwargs'])
            mdb.learn(
                from_data=data_source,
                to_predict=to_predict,
                **kwargs
            )

            stats = mindsdb_native.F.get_model_data(name)['data_analysis_v2']

            DatabaseWrapper(config).register_predictors([{
                'name': name,
                'predict': to_predict,
                'data_analysis': stats
            }], setup=False)

        if trx_type == 'predict':
            if isinstance(from_data, dict):
                when_data = from_data
            else:
                when_data = getattr(mindsdb_native, from_data['class'])(*from_data['args'], **from_data['kwargs'])

            predictions = mdb.predict(
                when_data=when_data,
                **kwargs
            )

            # @TODO Figure out a way to recover this since we are using `spawn` here... simple Queue or instiating a Multiprocessing manager and registering a value in a dict using that. Or using map from a multiprocessing pool with 1x process (though using a custom process there might be it's own bucket of annoying)
            return predictions



from mindsdb.integrations.clickhouse.clickhouse import Clickhouse
from mindsdb.integrations.postgres.postgres import PostgreSQL
from mindsdb.integrations.mariadb.mariadb import Mariadb
from mindsdb.integrations.mysql.mysql import MySQL
from mindsdb.integrations.mssql.mssql import MSSQL

from mindsdb.utilities.log import log as logger


class DatabaseWrapper():

    def __init__(self, config):
        self.config = config
        self._get_integrations()

    def _setup_integration(self, integration):
        success = False
        try:
            integration.setup()
            success = True
        except Exception as e:
            print('Failed to integrate with database ' + integration.name + f', error: {e}')
        return success

    def _get_integrations(self):
        # @TODO Once we have a presistent state sorted out this should be simplified as to not refresh the existing integrations every single time
        integration_arr = []
        for db_alias in self.config['integrations']:
            if self.config['integrations'][db_alias]['publish']:
                db_type = self.config['integrations'][db_alias]['type']
                if db_type == 'clickhouse':
                    integration_arr.append(Clickhouse(self.config, db_alias))
                elif db_type == 'mariadb':
                    integration_arr.append(Mariadb(self.config, db_alias))
                elif db_type == 'mysql':
                    integration_arr.append(MySQL(self.config, db_alias))
                elif db_type == 'postgres':
                    integration_arr.append(PostgreSQL(self.config, db_alias))
                elif db_type == 'mssql':
                    integration_arr.append(MSSQL(self.config, db_alias))
                elif db_type == 'mongodb':
                    pass
                else:
                    print(f'Uknown integration type: {db_type} for database called: {db_alias}')

        return integration_arr

    def register_predictors(self, model_data_arr, setup=True):
        it = self._get_integrations()
        for integration in it:
            register = True
            if setup:
                register = self._setup_integration(integration)
            if register:
                if integration.check_connection():
                    integration.register_predictors(model_data_arr)
                else:
                    logger.warning(f"There is no connection to {integration.name}. predictor wouldn't be registred.")

            integration = [integration]

    def unregister_predictor(self, name):
        for integration in self._get_integrations():
            if integration.check_connection():
                integration.unregister_predictor(name)
            else:
                logger.warning(f"There is no connection to {integration.name}. predictor wouldn't be unregistred")

    def check_connections(self):
        connections = {}
        for integration in self._get_integrations():
            connections[integration.name] = integration.check_connection()

        return connections



# Mindsdb native interface
from pathlib import Path
import json

from dateutil.parser import parse as parse_datetime

import mindsdb_native
from mindsdb_native import F
from mindsdb.utilities.fs import create_directory
from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
from mindsdb.interfaces.native.predictor_process import PredictorProcess
from mindsdb.interfaces.database.database import DatabaseWrapper


class MindsdbNative():
    def __init__(self, config):
        self.config = config
        self.dbw = DatabaseWrapper(self.config)

    def _setup_for_creation(self, name):
            predictor_dir = Path(self.config.paths['predictors']).joinpath(name)
            create_directory(predictor_dir)
            versions_file_path = predictor_dir.joinpath('versions.json')
            with open(str(versions_file_path), 'wt') as f:
                json.dump(self.config.versions, f, indent=4, sort_keys=True)

    def create(self, name):
        self._setup_for_creation(name)
        predictor = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})
        return predictor

    def learn(self, name, from_data, to_predict, kwargs={}):
        join_learn_process = kwargs.get('join_learn_process', False)
        if 'join_learn_process' in kwargs:
            del kwargs['join_learn_process']

        self._setup_for_creation(name)

        p = PredictorProcess(name, from_data, to_predict, kwargs, self.config.get_all(), 'learn')
        p.start()
        if join_learn_process is True:
            p.join()
            if p.exitcode != 0:
                raise Exception('Learning process failed !')

    def predict(self, name, when_data=None, kwargs={}):
        # @TODO Separate into two paths, one for "normal" predictions and one for "real time" predictions. Use the multiprocessing code commented out bellow for normal (once we figure out how to return the prediction object... else use the inline code but with the "real time" predict functionality of mindsdb_native taht will be implemented later)
        '''
        from_data = when if when is not None else when_data
        p = PredictorProcess(name, from_data, to_predict=None, kwargs=kwargs, config=self.config.get_all(), 'predict')
        p.start()
        predictions = p.join()
        '''
        mdb = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})

        predictions = mdb.predict(
            when_data=when_data,
            **kwargs
        )

        return predictions

    def analyse_dataset(self, ds):
        return F.analyse_dataset(ds)

    def get_model_data(self, name, native_view=False):
        model = F.get_model_data(name)
        if native_view:
            return model

        data_analysis = model['data_analysis_v2']
        for column in data_analysis['columns']:
            if len(data_analysis[column]) == 0 or data_analysis[column].get('empty', {}).get('is_empty', False):
                data_analysis[column]['typing'] = {
                    'data_subtype': DATA_SUBTYPES.INT
                }

        return model

    def get_models(self, status='any'):
        models = F.get_models()
        if status != 'any':
            models = [x for x in models if x['status'] == status]
        models = [x for x in models if x['status'] != 'training' or parse_datetime(x['created_at']) > parse_datetime(self.config['mindsdb_last_started_at'])]

        for i in range(len(models)):
            for k in ['train_end_at', 'updated_at', 'created_at']:
                if k in models[i] and models[i][k] is not None:
                    try:
                        models[i][k] = parse_datetime(str(models[i][k]).split('.')[0])
                    except Exception:
                        models[i][k] = parse_datetime(str(models[i][k]))
        return models

    def delete_model(self, name):
        F.delete_model(name)
        self.dbw.unregister_predictor(name)

    def rename_model(self, name, new_name):
        self.dbw.unregister_predictor(self.get_model_data(name))
        F.rename_model(name, new_name)
        self.dbw.register_predictors(self.get_model_data(new_name), setup=False)

    def load_model(self, fpath):
        F.import_model(model_archive_path=fpath)
        # @TODO How do we figure out the name here ?
        # dbw.register_predictors(...)

    def export_model(self, name):
        F.export_predictor(model_name=name)



from mindsdb.utilities.config import Config
from mindsdb.api.mongo.server import run_server
from mindsdb.utilities.log import initialize_log


def start(config, verbose=False):
    config = Config(config)
    if verbose:
        config['log']['level']['console'] = 'DEBUG'

    initialize_log(config, 'mongodb', wrap_print=True)

    run_server(config)



import os
import sys

from mindsdb.__about__ import __package_name__ as name, __version__   # noqa
from mindsdb.utilities.fs import get_or_create_dir_struct, create_dirs_recursive
from mindsdb.utilities.wizards import cli_config
from mindsdb.utilities.config import Config
from mindsdb.utilities.functions import args_parse, is_notebook

config_dir, storage_dir = get_or_create_dir_struct()

config_path = os.path.join(config_dir, 'config.json')
if not os.path.exists(config_path):
    _ = cli_config(None, None, storage_dir, config_dir, use_default=True)


try:
    if not is_notebook():
        args = args_parse()
        if args.config is not None:
            config_path = args.config
except:
    # This fials in some notebooks
    pass

try:
    mindsdb_config = Config(config_path)
except Exception as e:
    print(str(e))
    sys.exit(1)

paths = mindsdb_config.paths
create_dirs_recursive(paths)

os.environ['MINDSDB_STORAGE_PATH'] = paths['predictors']
os.environ['DEFAULT_LOG_LEVEL'] = os.environ.get('DEFAULT_LOG_LEVEL', 'ERROR')
os.environ['LIGHTWOOD_LOG_LEVEL'] = os.environ.get('LIGHTWOOD_LOG_LEVEL', 'ERROR')

from mindsdb_native import *
# Figure out how to add this as a module
import lightwood
#import dataskillet
import mindsdb.utilities.wizards as wizards
from mindsdb.interfaces.custom.model_interface import ModelInterface
